package com.sun.xml.internal.ws.model;

import org.xml.sax.SAXException;
import javax.xml.validation.SchemaFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLInputFactory;
import com.sun.xml.internal.ws.streaming.XMLStreamReaderUtil;
import javax.xml.bind.JAXBElement;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import com.sun.xml.internal.ws.util.xml.XmlUtil;
import javax.xml.bind.util.JAXBResult;
import javax.xml.bind.Unmarshaller;
import javax.xml.transform.Source;
import com.oracle.xmlns.internal.webservices.jaxws_databinding.ObjectFactory;
import java.net.URL;
import javax.xml.bind.JAXBContext;
import javax.xml.validation.Schema;
import org.w3c.dom.Element;
import com.oracle.xmlns.internal.webservices.jaxws_databinding.ExistingAnnotationsType;
import com.oracle.xmlns.internal.webservices.jaxws_databinding.JavaParam;
import com.oracle.xmlns.internal.webservices.jaxws_databinding.JavaMethod;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.lang.annotation.Annotation;
import java.io.FileNotFoundException;
import java.io.FileInputStream;
import javax.xml.transform.TransformerException;
import java.io.IOException;
import javax.xml.bind.JAXBException;
import java.util.Arrays;
import java.io.InputStream;
import javax.xml.transform.stream.StreamSource;
import java.util.Iterator;
import java.util.HashMap;
import java.io.File;
import java.util.Collection;
import com.oracle.xmlns.internal.webservices.jaxws_databinding.JavaWsdlMappingType;
import java.util.Map;

public class ExternalMetadataReader extends ReflectAnnotationReader
{
    private static final String NAMESPACE_WEBLOGIC_WSEE_DATABINDING = "http://xmlns.oracle.com/weblogic/weblogic-wsee-databinding";
    private static final String NAMESPACE_JAXWS_RI_EXTERNAL_METADATA = "http://xmlns.oracle.com/webservices/jaxws-databinding";
    private Map<String, JavaWsdlMappingType> readers;
    
    public ExternalMetadataReader(Collection<File> files, Collection<String> resourcePaths, ClassLoader classLoader, boolean xsdValidation, boolean disableXmlSecurity) {
        this.readers = new HashMap();
        if (files != null) {
            for (File file : files) {
                try {
                    String namespace = ExternalMetadataReader.Util.documentRootNamespace(this.newSource(file), disableXmlSecurity);
                    JavaWsdlMappingType externalMapping = this.parseMetadata(xsdValidation, this.newSource(file), namespace, disableXmlSecurity);
                    this.readers.put(externalMapping.getJavaTypeName(), externalMapping);
                }
                catch (Exception e) {
                    throw new RuntimeModelerException("runtime.modeler.external.metadata.unable.to.read", new Object[] { file.getAbsolutePath() });
                }
            }
        }
        if (resourcePaths != null) {
            for (String resourcePath : resourcePaths) {
                try {
                    String namespace = ExternalMetadataReader.Util.documentRootNamespace(this.newSource(resourcePath, classLoader), disableXmlSecurity);
                    JavaWsdlMappingType externalMapping = this.parseMetadata(xsdValidation, this.newSource(resourcePath, classLoader), namespace, disableXmlSecurity);
                    this.readers.put(externalMapping.getJavaTypeName(), externalMapping);
                }
                catch (Exception e) {
                    throw new RuntimeModelerException("runtime.modeler.external.metadata.unable.to.read", new Object[] { resourcePath });
                }
            }
        }
    }
    
    private StreamSource newSource(String resourcePath, ClassLoader classLoader) {
        InputStream is = classLoader.getResourceAsStream(resourcePath);
        return new StreamSource(is);
    }
    
    private JavaWsdlMappingType parseMetadata(boolean xsdValidation, StreamSource source, String namespace, boolean disableXmlSecurity) throws JAXBException, IOException, TransformerException {
        if ("http://xmlns.oracle.com/weblogic/weblogic-wsee-databinding".equals(namespace)) {
            return ExternalMetadataReader.Util.transformAndRead(source, disableXmlSecurity);
        }
        if ("http://xmlns.oracle.com/webservices/jaxws-databinding".equals(namespace)) {
            return ExternalMetadataReader.Util.read(source, xsdValidation, disableXmlSecurity);
        }
        throw new RuntimeModelerException("runtime.modeler.external.metadata.unsupported.schema", new Object[] { namespace, Arrays.asList(new String[] { "http://xmlns.oracle.com/weblogic/weblogic-wsee-databinding", "http://xmlns.oracle.com/webservices/jaxws-databinding" }).toString() });
    }
    
    private StreamSource newSource(File file) {
        try {
            return new StreamSource(new FileInputStream(file));
        }
        catch (FileNotFoundException e) {
            throw new RuntimeModelerException("runtime.modeler.external.metadata.unable.to.read", new Object[] { file.getAbsolutePath() });
        }
    }
    
    @Override
    public <A extends Annotation> A getAnnotation(Class<A> annType, Class<?> cls) {
        JavaWsdlMappingType r = this.reader(cls);
        return (r == null) ? super.getAnnotation(annType, cls) : ((A)ExternalMetadataReader.Util.annotation(r, annType));
    }
    
    private JavaWsdlMappingType reader(Class<?> cls) {
        return (JavaWsdlMappingType)this.readers.get(cls.getName());
    }
    
    Annotation[] getAnnotations(List<Object> objects) {
        ArrayList<Annotation> list = new ArrayList();
        for (Object a : objects) {
            if (Annotation.class.isInstance(a)) {
                list.add(Annotation.class.cast(a));
            }
            else {
                continue;
            }
        }
        return (Annotation[])list.toArray(new Annotation[list.size()]);
    }
    
    @Override
    public Annotation[] getAnnotations(final Class<?> c) {
        ExternalMetadataReader.Merger<Annotation[]> merger = new ExternalMetadataReader.Merger<Annotation[]>(this.reader(c)) {
            @Override
            Annotation[] reflection() {
                return ExternalMetadataReader.access$001(ExternalMetadataReader.this, c);
            }
            
            @Override
            Annotation[] external() {
                return ExternalMetadataReader.this.getAnnotations(this.reader.getClassAnnotation());
            }
        };
        return (Annotation[])merger.merge();
    }
    
    @Override
    public Annotation[] getAnnotations(final Method m) {
        ExternalMetadataReader.Merger<Annotation[]> merger = new ExternalMetadataReader.Merger<Annotation[]>(this.reader(m.getDeclaringClass())) {
            @Override
            Annotation[] reflection() {
                return ExternalMetadataReader.access$101(ExternalMetadataReader.this, m);
            }
            
            @Override
            Annotation[] external() {
                JavaMethod jm = ExternalMetadataReader.this.getJavaMethod(m, this.reader);
                return (jm == null) ? new Annotation[0] : ExternalMetadataReader.this.getAnnotations(jm.getMethodAnnotation());
            }
        };
        return (Annotation[])merger.merge();
    }
    
    @Override
    public <A extends Annotation> A getAnnotation(final Class<A> annType, final Method m) {
        ExternalMetadataReader.Merger<Annotation> merger = new ExternalMetadataReader.Merger<Annotation>(this.reader(m.getDeclaringClass())) {
            @Override
            Annotation reflection() {
                return ExternalMetadataReader.access$201(ExternalMetadataReader.this, annType, m);
            }
            
            @Override
            Annotation external() {
                JavaMethod jm = ExternalMetadataReader.this.getJavaMethod(m, this.reader);
                return (Annotation)ExternalMetadataReader.Util.annotation(jm, annType);
            }
        };
        return (A)merger.merge();
    }
    
    @Override
    public Annotation[][] getParameterAnnotations(final Method m) {
        ExternalMetadataReader.Merger<Annotation[][]> merger = new ExternalMetadataReader.Merger<Annotation[][]>(this.reader(m.getDeclaringClass())) {
            @Override
            Annotation[][] reflection() {
                return ExternalMetadataReader.access$301(ExternalMetadataReader.this, m);
            }
            
            @Override
            Annotation[][] external() {
                JavaMethod jm = ExternalMetadataReader.this.getJavaMethod(m, this.reader);
                Annotation[][] a = m.getParameterAnnotations();
                for (int i = 0; i < m.getParameterTypes().length; ++i) {
                    if (jm != null) {
                        JavaParam jp = (JavaParam)jm.getJavaParams().getJavaParam().get(i);
                        a[i] = ExternalMetadataReader.this.getAnnotations(jp.getParamAnnotation());
                    }
                }
                return a;
            }
        };
        return (Annotation[][])merger.merge();
    }
    
    @Override
    public void getProperties(Map<String, Object> prop, Class<?> cls) {
        JavaWsdlMappingType r = this.reader(cls);
        if (r == null || ExistingAnnotationsType.MERGE.equals(r.getExistingAnnotations())) {
            super.getProperties(prop, cls);
        }
    }
    
    @Override
    public void getProperties(Map<String, Object> prop, Method m) {
        JavaWsdlMappingType r = this.reader(m.getDeclaringClass());
        if (r == null || ExistingAnnotationsType.MERGE.equals(r.getExistingAnnotations())) {
            super.getProperties(prop, m);
        }
        if (r != null) {
            JavaMethod jm = this.getJavaMethod(m, r);
            Element[] e = ExternalMetadataReader.Util.annotation(jm);
            prop.put("eclipselink-oxm-xml.xml-element", this.findXmlElement(e));
        }
    }
    
    @Override
    public void getProperties(Map<String, Object> prop, Method m, int pos) {
        JavaWsdlMappingType r = this.reader(m.getDeclaringClass());
        if (r == null || ExistingAnnotationsType.MERGE.equals(r.getExistingAnnotations())) {
            super.getProperties(prop, m, pos);
        }
        if (r != null) {
            JavaMethod jm = this.getJavaMethod(m, r);
            if (jm == null) {
                return;
            }
            JavaParam jp = (JavaParam)jm.getJavaParams().getJavaParam().get(pos);
            Element[] e = ExternalMetadataReader.Util.annotation(jp);
            prop.put("eclipselink-oxm-xml.xml-element", this.findXmlElement(e));
        }
    }
    
    JavaMethod getJavaMethod(Method method, JavaWsdlMappingType r) {
        JavaWsdlMappingType.JavaMethods javaMethods = r.getJavaMethods();
        if (javaMethods == null) {
            return null;
        }
        List<JavaMethod> sameName = new ArrayList();
        for (JavaMethod jm : javaMethods.getJavaMethod()) {
            if (method.getName().equals(jm.getName())) {
                sameName.add(jm);
            }
            else {
                continue;
            }
        }
        if (sameName.isEmpty()) {
            return null;
        }
        if (sameName.size() == 1) {
            return (JavaMethod)sameName.get(0);
        }
        Class<?>[] argCls = method.getParameterTypes();
        for (JavaMethod jm2 : sameName) {
            JavaMethod.JavaParams params = jm2.getJavaParams();
            if (params != null && params.getJavaParam() != null && params.getJavaParam().size() == argCls.length) {
                int count = 0;
                for (int i = 0; i < argCls.length; ++i) {
                    JavaParam jp = (JavaParam)params.getJavaParam().get(i);
                    if (argCls[i].getName().equals(jp.getJavaType())) {
                        ++count;
                    }
                }
                if (count == argCls.length) {
                    return jm2;
                }
                continue;
            }
            else {
                continue;
            }
        }
        return null;
    }
    
    Element findXmlElement(Element[] xa) {
        if (xa == null) {
            return null;
        }
        int length = xa.length;
        int i = 0;
        while (i < length) {
            Element e = xa[i];
            if (e.getLocalName().equals("java-type")) {
                return e;
            }
            if (e.getLocalName().equals("xml-element")) {
                return e;
            }
            ++i;
        }
        return null;
    }
    
    static Annotation[] access$001(ExternalMetadataReader x0, Class x1) {
        return x0.getAnnotations(x1);
    }
    
    static Annotation[] access$101(ExternalMetadataReader x0, Method x1) {
        return x0.getAnnotations(x1);
    }
    
    static Annotation access$201(ExternalMetadataReader x0, Class x1, Method x2) {
        return x0.getAnnotation(x1, x2);
    }
    
    static Annotation[][] access$301(ExternalMetadataReader x0, Method x1) {
        return x0.getParameterAnnotations(x1);
    }
    
    abstract static class Merger<T>
    {
        JavaWsdlMappingType reader;
        
        Merger(JavaWsdlMappingType r) {
            this.reader = r;
        }
        
        abstract T reflection();
        
        abstract T external();
        
        T merge() {
            T reflection = this.reflection();
            if (this.reader == null) {
                return reflection;
            }
            T external = this.external();
            if (!ExistingAnnotationsType.MERGE.equals(this.reader.getExistingAnnotations())) {
                return external;
            }
            if ((reflection instanceof Annotation)) {
                return this.doMerge((Annotation)reflection, (Annotation)external);
            }
            if ((reflection instanceof Annotation[][])) {
                return this.doMerge((Annotation[][])(Annotation[][])reflection, (Annotation[][])(Annotation[][])external);
            }
            return this.doMerge((Annotation[])(Annotation[])reflection, (Annotation[])(Annotation[])external);
        }
        
        private Annotation doMerge(Annotation reflection, Annotation external) {
            return (external != null) ? external : reflection;
        }
        
        private Annotation[][] doMerge(Annotation[][] reflection, Annotation[][] external) {
            for (int i = 0; i < reflection.length; ++i) {
                reflection[i] = this.doMerge(reflection[i], (external.length > i) ? external[i] : null);
            }
            return reflection;
        }
        
        private Annotation[] doMerge(Annotation[] annotations, Annotation[] externalAnnotations) {
            HashMap<String, Annotation> mergeMap = new HashMap();
            if (annotations != null) {
                for (Annotation reflectionAnnotation : annotations) {
                    mergeMap.put(reflectionAnnotation.annotationType().getName(), reflectionAnnotation);
                }
            }
            if (externalAnnotations != null) {
                for (Annotation externalAnnotation : externalAnnotations) {
                    mergeMap.put(externalAnnotation.annotationType().getName(), externalAnnotation);
                }
            }
            Collection<Annotation> values = mergeMap.values();
            int size = values.size();
            return (size == 0) ? null : ((Annotation[])values.toArray(new Annotation[size]));
        }
    }
    
    static class Util
    {
        private static final String DATABINDING_XSD = "jaxws-databinding.xsd";
        private static final String TRANSLATE_NAMESPACES_XSL = "jaxws-databinding-translate-namespaces.xml";
        static Schema schema;
        static JAXBContext jaxbContext;
        
        private static URL getResource() {
            ClassLoader classLoader = ExternalMetadataReader.Util.class.getClassLoader();
            return (classLoader != null) ? classLoader.getResource("jaxws-databinding.xsd") : ClassLoader.getSystemResource("jaxws-databinding.xsd");
        }
        
        private static JAXBContext createJaxbContext(boolean disableXmlSecurity) {
            Class[] cls = { ObjectFactory.class };
            try {
                if (disableXmlSecurity) {
                    Map<String, Object> properties = new HashMap();
                    properties.put("com.sun.xml.internal.bind.disableXmlSecurity", Boolean.valueOf(disableXmlSecurity));
                    return JAXBContext.newInstance(cls, properties);
                }
                return JAXBContext.newInstance(cls);
            }
            catch (JAXBException e) {
                e.printStackTrace();
                return null;
            }
        }
        
        public static JavaWsdlMappingType read(Source src, boolean xsdValidation, boolean disableXmlSecurity) throws IOException, JAXBException {
            JAXBContext ctx = jaxbContext(disableXmlSecurity);
            try {
                Unmarshaller um = ctx.createUnmarshaller();
                if (xsdValidation) {
                    if (ExternalMetadataReader.Util.schema != null) {}
                    um.setSchema(ExternalMetadataReader.Util.schema);
                }
                Object o = um.unmarshal(src);
                return getJavaWsdlMapping(o);
            }
            catch (JAXBException e) {
                URL url = new URL(src.getSystemId());
                Source s = new StreamSource(url.openStream());
                Unmarshaller um2 = ctx.createUnmarshaller();
                if (xsdValidation) {
                    if (ExternalMetadataReader.Util.schema != null) {}
                    um2.setSchema(ExternalMetadataReader.Util.schema);
                }
                Object o2 = um2.unmarshal(s);
                return getJavaWsdlMapping(o2);
            }
        }
        
        private static JAXBContext jaxbContext(boolean disableXmlSecurity) {
            return disableXmlSecurity ? createJaxbContext(true) : ExternalMetadataReader.Util.jaxbContext;
        }
        
        public static JavaWsdlMappingType transformAndRead(Source src, boolean disableXmlSecurity) throws TransformerException, JAXBException {
            Source xsl = new StreamSource(ExternalMetadataReader.Util.class.getResourceAsStream("jaxws-databinding-translate-namespaces.xml"));
            JAXBResult result = new JAXBResult(jaxbContext(disableXmlSecurity));
            TransformerFactory tf = XmlUtil.newTransformerFactory(!disableXmlSecurity);
            Transformer transformer = tf.newTemplates(xsl).newTransformer();
            transformer.transform(src, result);
            return getJavaWsdlMapping(result.getResult());
        }
        
        static JavaWsdlMappingType getJavaWsdlMapping(Object o) {
            Object val = (o instanceof JAXBElement) ? ((JAXBElement)o).getValue() : o;
            if ((val instanceof JavaWsdlMappingType)) {
                return (JavaWsdlMappingType)val;
            }
            return null;
        }
        
        static <T> T findInstanceOf(Class<T> type, List<Object> objects) {
            for (Object o : objects) {
                if (type.isInstance(o)) {
                    return type.cast(o);
                }
                continue;
            }
            return null;
        }
        
        public static <T> T annotation(JavaWsdlMappingType jwse, Class<T> anntype) {
            if (jwse == null || jwse.getClassAnnotation() == null) {
                return null;
            }
            return findInstanceOf(anntype, jwse.getClassAnnotation());
        }
        
        public static <T> T annotation(JavaMethod jm, Class<T> anntype) {
            if (jm == null || jm.getMethodAnnotation() == null) {
                return null;
            }
            return findInstanceOf(anntype, jm.getMethodAnnotation());
        }
        
        public static <T> T annotation(JavaParam jp, Class<T> anntype) {
            if (jp == null || jp.getParamAnnotation() == null) {
                return null;
            }
            return findInstanceOf(anntype, jp.getParamAnnotation());
        }
        
        public static Element[] annotation(JavaMethod jm) {
            if (jm == null || jm.getMethodAnnotation() == null) {
                return null;
            }
            return findElements(jm.getMethodAnnotation());
        }
        
        public static Element[] annotation(JavaParam jp) {
            if (jp == null || jp.getParamAnnotation() == null) {
                return null;
            }
            return findElements(jp.getParamAnnotation());
        }
        
        private static Element[] findElements(List<Object> objects) {
            List<Element> elems = new ArrayList();
            for (Object o : objects) {
                if ((o instanceof Element)) {
                    elems.add((Element)o);
                }
                else {
                    continue;
                }
            }
            return (Element[])elems.toArray(new Element[elems.size()]);
        }
        
        static String documentRootNamespace(Source src, boolean disableXmlSecurity) throws XMLStreamException {
            XMLInputFactory factory = XmlUtil.newXMLInputFactory(!disableXmlSecurity);
            XMLStreamReader streamReader = factory.createXMLStreamReader(src);
            XMLStreamReaderUtil.nextElementContent(streamReader);
            String namespaceURI = streamReader.getName().getNamespaceURI();
            XMLStreamReaderUtil.close(streamReader);
            return namespaceURI;
        }
        
        static {
            SchemaFactory sf = SchemaFactory.newInstance("http://www.w3.org/2001/XMLSchema");
            try {
                URL xsdUrl = getResource();
                if (xsdUrl != null) {
                    ExternalMetadataReader.Util.schema = sf.newSchema(xsdUrl);
                }
            }
            catch (SAXException ex) {}
            ExternalMetadataReader.Util.jaxbContext = createJaxbContext(false);
        }
    }
}
